﻿/*
 * MazeGenerator 1.0
 * 
 * Descriptions of the used generation algorithms can be found here:
 * http://weblog.jamisbuck.org/2011/2/7/maze-generation-algorithm-recap
 * http://www.mazeworks.com/mazegen/index.htm
 * 
 * www.sunshine2k.de || www.bastian-molkenthin.de
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using MazeGenerator.MazeGenerator;
using MazeGenerator.MazeGenerator.StepwiseMazeGeneratorAlgorithm;
using MazeGenerator.MazeGenerator.MazeGeneratorAlgorithm;
using System.Windows.Automation.Peers;
using System.Windows.Automation.Provider;
using System.Windows.Threading;
using System.Text.RegularExpressions;

namespace MazeGenerator
{
    public partial class MainPage : UserControl
    {
        private static string DIRECTGEN_BUTTON_TEXT = "Generate";
        private static string STEPWISEGEN_BUTTON_TEXT = "Generate Stepwise";
        
        Maze maze;
        MazeDrawer mazeDrawer;

        /// <summary>
        /// current immediate maze generation object
        /// </summary>
        AbstractImmediateMazeGenerator mazeGenImmediate;

        /// <summary>
        /// current stepwise maze generation object
        /// </summary>
        AbstractStepwiseMazeGenerator mazeGenStepwise;

        /// <summary>
        /// Flag if stepwise maze generation is running or not.
        /// </summary>
        bool mazeGenStepwiseFinished;

        /// <summary>
        /// Holds the size in cells of the maze.
        /// </summary>
        int mazeSize;

        public MainPage()
        {
            InitializeComponent();
            UpdateLayout();

            // create initial 10x10 grid just to show something
            maze = new Maze(10, 10);
            mazeSize = 10;
            // init drawer class to visualize grid
            mazeDrawer = new MazeDrawer(maze, gridCanvas);
            
            /* set default layout */
            radioButtonDfsRecursive.IsChecked = true;
            radioButtonDirectGen.IsChecked = true;
            genButton.Content = DIRECTGEN_BUTTON_TEXT;

            slider1.ValueChanged += new RoutedPropertyChangedEventHandler<double>(slider1_ValueChanged);

            //System.Diagnostics.Debug.WriteLine("Test");
        }

        /// <summary>
        /// Called when the generate button is clicked.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void genButton_Click(object sender, RoutedEventArgs e)
        {
            if (genButton.Content.Equals(DIRECTGEN_BUTTON_TEXT))
            {
                maze = null;
                mazeGenImmediate = null;
                mazeDrawer.Dispose();
                mazeDrawer = null;
                System.GC.Collect();

                mazeGenImmediate = GetDirectMazeGenerator();
                mazeGenImmediate.Generate();
                mazeDrawer = new MazeDrawer(mazeGenImmediate, gridCanvas);
            }
            else if (genButton.Content.Equals(STEPWISEGEN_BUTTON_TEXT))
            {
                if (mazeGenStepwise == null || mazeGenStepwiseFinished)
                {   /* no stepwise generation running, start new one */
                    if (mazeGenStepwise != null)
                    {
                        mazeGenStepwise.Dispose();
                    }
                    mazeGenStepwise = null;
                    mazeDrawer.Dispose();
                    mazeDrawer = null;
                    System.GC.Collect();
                    
                    mazeGenStepwiseFinished = false;
                    mazeGenStepwise = GetStepwiseMazeGenerator();
                    mazeDrawer = new MazeDrawer(mazeGenStepwise, gridCanvas);
                    mazeGenStepwise.MazeChanged += new EventHandler<MazeGenerator.StepwiseMazeGeneratorAlgorithm.MazeChangedEvent>(algo_MazeChanged);
                    mazeGenStepwise.Start();
                }
                else
                {   /* calculate next maze step */
                    mazeGenStepwise.Continue();
                }
                
            }
        }

        /// <summary>
        /// Get a new instance of a stepwise maze generator depending on the GUI selection.
        /// </summary>
        /// <returns></returns>
        private AbstractStepwiseMazeGenerator GetStepwiseMazeGenerator()
        {
            Tuple<int, int> startCell = GetStartCellTuple();
            if (radioButtonDfsRecursive.IsChecked.Value)
            {
                return new StepwiseMazeGenDFS(mazeSize, mazeSize, this, startCell.Item1, startCell.Item2);
            }
            else if (radioButtonHuntAndKill.IsChecked.Value)
            {
                return new StepwiseMazeGenHuntAndKill(mazeSize, mazeSize, this, startCell.Item1, startCell.Item2);
            }
            else if (radioButtonEllerAlgo.IsChecked.Value)
            {
                return new StepwiseMazeGenEller(mazeSize, mazeSize, this, startCell.Item1, startCell.Item2);
            }
            else
            {
                throw new InvalidProgramException();
            }
        }

        /// <summary>
        /// Get a new instance of a immediate maze generator depending on the GUI selection.
        /// </summary>
        /// <returns></returns>
        private AbstractImmediateMazeGenerator GetDirectMazeGenerator()
        {
            Tuple<int, int> startCell = GetStartCellTuple();
            if (radioButtonDfsRecursive.IsChecked.Value)
            {
                return new MazeGenDFSRecursive(mazeSize, mazeSize, startCell.Item1, startCell.Item2);
            }
            else if (radioButtonDfsIterative.IsChecked.Value)
            {
                return new MazeGenDFS(mazeSize, mazeSize, startCell.Item1, startCell.Item2);
            }
            else if (radioButtonHuntAndKill.IsChecked.Value)
            {
                return new MazeGenHuntAndKill(mazeSize, mazeSize, startCell.Item1, startCell.Item2);
            }
            else if (radioButtonEllerAlgo.IsChecked.Value)
            {
                return new MazeGenEller(mazeSize, mazeSize, startCell.Item1, startCell.Item2);
            }
            else
            {
                throw new InvalidProgramException();
            }
        }

        /// <summary>
        /// Called when the maze has changed during stepwise calculation.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void algo_MazeChanged(object sender, MazeChangedEvent e)
        {
            mazeGenStepwiseFinished = e.IsFinished;
            mazeDrawer.updateGrid(mazeGenStepwise);
            if (e.IsFinished)
            {
                genButton.IsEnabled = false;
            }
        }

        /// <summary>
        /// Called when a algorithm calculation type is selected.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void radioButtonGenType_Checked(object sender, RoutedEventArgs e)
        {
            if (radioButtonDirectGen.IsChecked.Value)
            {
                resetButton.IsEnabled = false;
                radioButtonDfsIterative.IsEnabled = true;
                genButton.Content = DIRECTGEN_BUTTON_TEXT;
            }
            else if (radioButtonStepwiseGen.IsChecked.Value)
            {
                resetButton.IsEnabled = true;
                radioButtonDfsIterative.IsEnabled = false;
                genButton.Content = STEPWISEGEN_BUTTON_TEXT;
            }
            else
            {
                throw new InvalidProgramException();
            }
        }

        /// <summary>
        /// Called when the reset button is clicked.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void resetButton_Click(object sender, RoutedEventArgs e)
        {
            mazeGenStepwise = null;
            genButton.IsEnabled = true;
            /* simulate button click */
            ButtonAutomationPeer peer = new ButtonAutomationPeer(genButton);
            IInvokeProvider invokeProv = peer.GetPattern(PatternInterface.Invoke) as IInvokeProvider;
            invokeProv.Invoke();
        }

        /// <summary>
        /// Called when the maze size slider changed.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void slider1_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            mazeSizeText.Text = "Size: " + (int)e.NewValue;
            mazeSize = (int)e.NewValue;
            if (resetButton.IsEnabled)
            {
                ButtonAutomationPeer peer = new ButtonAutomationPeer(resetButton);
                IInvokeProvider invokeProv = peer.GetPattern(PatternInterface.Invoke) as IInvokeProvider;
                invokeProv.Invoke();
            }
        }

        private Tuple<int, int> GetStartCellTuple()
        {
            if (Regex.IsMatch(startCellXTextBox.Text, "^[0-9 ]+$") &&
                Regex.IsMatch(startCellYTextBox.Text, "^[0-9 ]+$"))
            {
                return new Tuple<int, int>(int.Parse(startCellXTextBox.Text), int.Parse(startCellYTextBox.Text));
            }
            else
            {
                return new Tuple<int, int>(0, 0);
            }
        }

        /// <summary>
        /// TEST
        /// </summary>
        AbstractImmediateMazeGenerator mymaze = null;
        private void button1_Click(object sender, RoutedEventArgs e)
        {
            mymaze = new MazeGenEller(8, 8);

            mymaze.Generate();
            mazeDrawer = new MazeDrawer(mymaze, gridCanvas);
        }

    }
}
