﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using MazeGenerator.MazeGenerator;
using System.Collections.Generic;

namespace MazeGenerator
{
    /// <summary>
    /// Graphical representation of one maze.
    /// </summary>
    public class MazeDrawer : IDisposable
    {
        /// <summary>
        /// reference to maze to draw
        /// </summary>
        private Maze maze;
        
        /// <summary>
        /// parent canvas to draw on
        /// </summary>
        private Canvas mazeCanvas;

        /// <summary>
        /// array holding the cells of the maze
        /// </summary>
        private MazeDrawerCell[,] drawCells;

        /// <summary>
        /// margin between canvas border and maze borders.
        /// </summary>1
        private static int CONST_MARGIN = 10;

        private Brush startCellFillBrush = new SolidColorBrush(Color.FromArgb(255, 255, 200, 200));
        private Brush lastCellFillBrush = new SolidColorBrush(Colors.Orange);
        private Brush visitedCellFillBrush = new SolidColorBrush(Color.FromArgb(255, 220, 220, 220));
        private Brush untouchedCellFillBrush = new SolidColorBrush(Colors.White);

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="maze">maze to draw</param>
        /// <param name="mazeCanvas">canvas to draw maze on</param>
        public MazeDrawer(Maze maze, Canvas mazeCanvas)
        {
            this.maze = maze;
            this.mazeCanvas = mazeCanvas;

            this.mazeCanvas.LayoutUpdated += new EventHandler(mazeCanvas_LayoutUpdated);
            buildGridStructure();
        }

        /// <summary>
        /// Called when parent canvas resizes. -> Also maze shall be redrawn.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void mazeCanvas_LayoutUpdated(object sender, EventArgs e)
        {
            int drawCellWidth, drawCellHeight;

            drawCellWidth = (int)((mazeCanvas.ActualWidth - 2 * CONST_MARGIN) / maze.Width);
            drawCellHeight = (int)((mazeCanvas.ActualHeight - 2 * CONST_MARGIN) / maze.Height);

            for (int x = 0; x < maze.Width; x++)
            {
                for (int y = 0; y < maze.Height; y++)
                {
                    // calculate position of cell vertices
                    int xLeft = CONST_MARGIN + x * drawCellWidth;
                    int xRight = CONST_MARGIN + (x + 1) * drawCellWidth;
                    int yTop = CONST_MARGIN + y * drawCellHeight;
                    int yBottom = CONST_MARGIN + (y + 1) * drawCellHeight;

                    // set the position of all edges and cell
                    drawCells[x, y].SetEdgePos(MazeDir.NORTH, xLeft, yTop, xRight, yTop);
                    drawCells[x, y].SetEdgePos(MazeDir.EAST, xRight, yTop, xRight, yBottom);
                    drawCells[x, y].SetEdgePos(MazeDir.SOUTH, xLeft, yBottom, xRight, yBottom);
                    drawCells[x, y].SetEdgePos(MazeDir.WEST, xLeft, yTop, xLeft, yBottom);
                    drawCells[x, y].SetCellRectanglePos(xLeft, yTop, drawCellWidth, drawCellHeight);

                    // color cell depending on state
                    if (maze.IsCellStartCell(x, y))
                    {
                        drawCells[x, y].GetCellRectangle().Fill = startCellFillBrush;
                    }
                    else if (maze.IsCellLastVisited(x, y))
                    {
                        drawCells[x, y].GetCellRectangle().Fill = lastCellFillBrush;
                    }
                    else if (maze.IsCellVisited(x, y))
                    {
                        drawCells[x, y].GetCellRectangle().Fill = visitedCellFillBrush;
                    }
                    else
                    {
                        drawCells[x, y].GetCellRectangle().Fill = untouchedCellFillBrush;
                    }

                    // in case edge was removed, color it dso that it has the same color as the cell
                    foreach (MazeDir dir in MazeDir.NORTH.GetAllDir())
                    {
                        if (!maze.ExistsWall(x, y, dir))
                        {
                            drawCells[x, y].GetEdge(dir).GetEdgeLine().Stroke = visitedCellFillBrush;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Build the initial grid and add the edge lines to the parent canvas.
        /// </summary>
        private void buildGridStructure()
        {
            mazeCanvas.Children.Clear();

            drawCells = new MazeDrawerCell[maze.Width, maze.Height];
            // create topleft cell
            drawCells[0, 0] = new MazeDrawerCell();
            mazeCanvas.Children.Add(drawCells[0, 0].GetCellRectangle());
            AddEdgeToDrawCell(drawCells[0, 0], MazeDir.NORTH, new MazeDrawerEdge(), true);
            AddEdgeToDrawCell(drawCells[0, 0], MazeDir.EAST, new MazeDrawerEdge(), true);
            AddEdgeToDrawCell(drawCells[0, 0], MazeDir.SOUTH, new MazeDrawerEdge(), true);
            AddEdgeToDrawCell(drawCells[0, 0], MazeDir.WEST, new MazeDrawerEdge(), true);

            // create rest of first row of cells
		    for (int x = 1; x < maze.Width; x++)
		    {		
			    drawCells[x, 0] = new MazeDrawerCell();
                mazeCanvas.Children.Add(drawCells[x, 0].GetCellRectangle());
                AddEdgeToDrawCell(drawCells[x, 0], MazeDir.NORTH, new MazeDrawerEdge(), true);
                AddEdgeToDrawCell(drawCells[x, 0], MazeDir.EAST, new MazeDrawerEdge(), true);
                AddEdgeToDrawCell(drawCells[x, 0], MazeDir.SOUTH, new MazeDrawerEdge(), true);
                AddEdgeToDrawCell(drawCells[x, 0], MazeDir.WEST, drawCells[x - 1, 0].GetEdge(MazeDir.EAST), false);
		    }

            for (int y = 1; y < maze.Height; y++)
            {
                // create first cell in column
                drawCells[0, y] = new MazeDrawerCell();
                mazeCanvas.Children.Add(drawCells[0, y].GetCellRectangle());
                AddEdgeToDrawCell(drawCells[0, y], MazeDir.NORTH, drawCells[0, y - 1].GetEdge(MazeDir.SOUTH), false);
                AddEdgeToDrawCell(drawCells[0, y], MazeDir.EAST, new MazeDrawerEdge(), true);
                AddEdgeToDrawCell(drawCells[0, y], MazeDir.SOUTH, new MazeDrawerEdge(), true);
                AddEdgeToDrawCell(drawCells[0, y], MazeDir.WEST, new MazeDrawerEdge(), true);

                // create rest of this row
                for (int x = 1; x < maze.Width; x++)
                {
                    drawCells[x, y] = new MazeDrawerCell();
                    mazeCanvas.Children.Add(drawCells[x, y].GetCellRectangle());
                    AddEdgeToDrawCell(drawCells[x, y], MazeDir.NORTH, drawCells[x, y - 1].GetEdge(MazeDir.SOUTH), false);
                    AddEdgeToDrawCell(drawCells[x, y], MazeDir.EAST, new MazeDrawerEdge(), true);
                    AddEdgeToDrawCell(drawCells[x, y], MazeDir.SOUTH, new MazeDrawerEdge(), true);
                    AddEdgeToDrawCell(drawCells[x, y], MazeDir.WEST, drawCells[x - 1, y].GetEdge(MazeDir.EAST), false);
                }
            }
        }

        /// <summary>
        /// Add an edge to a cell side. Optionally adds the edge also to the parent canvas.
        /// </summary>
        /// <param name="cell"></param>
        /// <param name="dir"></param>
        /// <param name="edge"></param>
        /// <param name="bAddToCanvas"></param>
        private void AddEdgeToDrawCell(MazeDrawerCell cell, MazeDir dir, MazeDrawerEdge edge, bool bAddToCanvas)
        {
            cell.SetEdge(dir, edge);
            if (bAddToCanvas)
            {
                mazeCanvas.Children.Add(edge.GetEdgeLine());
            }
        }

        /// <summary>
        /// Update the graphical maze with the given maze data.
        /// </summary>
        /// <param name="maze"></param>
        public void updateGrid(Maze maze)
        {
            this.maze = maze;
            buildGridStructure();
        }

        /// <summary>
        /// Clean up.
        /// </summary>
        public void Dispose()
        {
            this.mazeCanvas.LayoutUpdated -= new EventHandler(mazeCanvas_LayoutUpdated);
        }
    }

}
