﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace MazeGenerator.MazeGenerator
{
    public class Maze
    {
        public int Width { get; private set; }

        public int Height { get; private set; }

        private ushort[,] cells;

        /// <summary>
        /// x and y cell coordinate of last visited cell
        /// </summary>
        private int lastCellX, lastCellY;

        /// <summary>
        /// x and y cell coordinate of start cell
        /// </summary>
        protected int StartCellX { get; private set; }
        protected int StartCellY { get; private set; }

        private static readonly ushort WALL_NORTH   = (1 << (ushort)MazeDir.NORTH); // 0x0001
        private static readonly ushort WALL_EAST    = (1 << (ushort)MazeDir.EAST);  // 0x0002
        private static readonly ushort WALL_SOUTH   = (1 << (ushort)MazeDir.SOUTH); // 0x0004
        private static readonly ushort WALL_WEST    = (1 << (ushort)MazeDir.WEST);  // 0x0008

        private static readonly ushort CELL_VISITED = 0x8000;

        /* order is: NORTH (0), EAST (1), SOUTH (2), WEST (3) */
        protected static readonly int[] DELTA_X = { 0, 1, 0,-1 };
        protected static readonly int[] DELTA_Y = {-1, 0, 1, 0 };
        protected static readonly MazeDir[] ADJACENTSIDE = { MazeDir.SOUTH, MazeDir.WEST, MazeDir.NORTH, MazeDir.EAST };

        /// <summary>
        /// Create a new maze with given number of cells
        /// </summary>
        /// <param name="width"></param>
        /// <param name="height"></param>
        public Maze(int width, int height, int startCellX, int startCellY)
        {
            Width = width;
            Height = height;
            StartCellX = startCellX;
            StartCellY = startCellY;
            lastCellX = lastCellY = -1;
            cells = new ushort[width, height];
            for (int cx = 0; cx < width; cx++)
            {
                for (int cy = 0; cy < height; cy++)
                {
                    cells[cx, cy] |= (ushort)(0x0000 | (ushort)(WALL_NORTH | WALL_EAST | WALL_SOUTH | WALL_WEST));
                }
            }
        }

        public Maze(int width, int height) : this(width, height, -1, -1) { }

        /// <summary>
        /// Set a wall/edge of a maze cell.
        /// </summary>
        /// <param name="cx"></param>
        /// <param name="cy"></param>
        /// <param name="dir"></param>
        public void SetWall(int cx, int cy, MazeDir dir)
        {
            Wall_Set(cx, cy, dir);
            
            // add the same wall of the adjacent wall if exist
            int adjCellX = cx + DELTA_X[(ushort)dir];
            int adjCellY = cy + DELTA_Y[(ushort)dir];
            if ((adjCellX >= 0) && (adjCellX < Width) && (adjCellY >= 0) && (adjCellY < Height))
            {
                Wall_Set(adjCellX, adjCellY, ADJACENTSIDE[(ushort)dir]);
            }
        }

        /// <summary>
        /// Remove a wall/edge from a maze cell
        /// </summary>
        /// <param name="cx"></param>
        /// <param name="cy"></param>
        /// <param name="dir"></param>
        public void RemoveWall(int cx, int cy, MazeDir dir)
        {
            Wall_Remove(cx, cy, dir);

            // remove the same wall of the adjacent wall if exist
            int adjCellX = cx + DELTA_X[(ushort)dir];
            int adjCellY = cy + DELTA_Y[(ushort)dir];
            if ((adjCellX >= 0) && (adjCellX < Width) && (adjCellY >= 0) && (adjCellY < Height))
            {
                Wall_Remove(adjCellX, adjCellY, ADJACENTSIDE[(ushort)dir]);
            }
        }

        /// <summary>
        /// Checks if a wall/edge in given directory exists.
        /// A wall on the edge grid exists always.
        /// </summary>
        /// <param name="cx"></param>
        /// <param name="cy"></param>
        /// <param name="dir"></param>
        /// <returns></returns>
        public bool ExistsWall(int cx, int cy, MazeDir dir)
        {
            int adjCellX = cx + DELTA_X[(ushort)dir];
            int adjCellY = cy + DELTA_Y[(ushort)dir];
            if ((adjCellX >= 0) && (adjCellX < Width) && (adjCellY >= 0) && (adjCellY < Height))
            {
                return ((cells[cx, cy] & (1 << (ushort)dir)) != 0);
            }
            else
            {   /* outside border is interpreted as wall */
                return true;
            }
        }

        /// <summary>
        /// Mark a cell as already handled.
        /// </summary>
        /// <param name="cx"></param>
        /// <param name="cy"></param>
        public void VisitCell(int cx, int cy)
        {
            cells[cx, cy] |= CELL_VISITED;
            lastCellX = cx;
            lastCellY = cy;
        }

        /// <summary>
        /// Check if given cell was already visited.
        /// </summary>
        /// <param name="cx"></param>
        /// <param name="cy"></param>
        /// <returns></returns>
        public bool IsCellVisited(int cx, int cy)
        {
            return ((cells[cx, cy] & CELL_VISITED) != 0);
        }

        /// <summary>
        /// Check if cell in direction of given cell was already visited.
        /// </summary>
        /// <param name="cx"></param>
        /// <param name="cy"></param>
        /// <param name="dir"></param>
        /// <returns></returns>
        public bool IsAdjacentCellVisited(int cx, int cy, MazeDir dir)
        {
            int adjCellX = cx + DELTA_X[(ushort)dir];
            int adjCellY = cy + DELTA_Y[(ushort)dir];
            if ((adjCellX >= 0) && (adjCellX < Width) && (adjCellY >= 0) && (adjCellY < Height))
            {
                return IsCellVisited(adjCellX, adjCellY);
            }
            else
            {
                // cell does not exist, return TRUE
                return true;
            }
        }

        /// <summary>
        /// Check wether the given cell position is valid, e.g. identifies a cell inside the maze.
        /// </summary>
        /// <param name="cx"></param>
        /// <param name="cy"></param>
        /// <returns></returns>
        public bool IsCellInsideMaze(int cx, int cy)
        {
            return (cx >= 0 && cy >= 0 && cx < Width && cy < Height);
        }

        /// <summary>
        /// Check wether the adjacent cell of given cell position is valid, 
        /// e.g. identifies a cell inside the maze.
        /// </summary>
        /// <param name="cx"></param>
        /// <param name="cy"></param>
        /// <param name="dir"></param>
        /// <returns></returns>
        public bool IsAdjacentCellInsideMaze(int cx, int cy, MazeDir dir)
        {
            int adjCellX = cx + DELTA_X[(ushort)dir];
            int adjCellY = cy + DELTA_Y[(ushort)dir];
            return (adjCellX >= 0 && adjCellY >= 0 && adjCellX < Width && adjCellY < Height);
        }

        public bool IsCellLastVisited(int cx, int cy)
        {
            return ((cx == lastCellX) && (cy == lastCellY));
        }

        public bool IsCellStartCell(int cx, int cy)
        {
            return ((cx == StartCellX) && (cy == StartCellY));
        }

        /************************************
         * Private helper functions
         ************************************/

        private void Wall_Set(int cx, int cy, MazeDir dir)
        {
            cells[cx, cy] |= (ushort)(1 << (ushort)dir);
        }

        private void Wall_Remove(int cx, int cy, MazeDir dir)
        {
            cells[cx, cy] &= (ushort)~(1 << (ushort)dir);
        }
        
    }


}
