﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Collections.Generic;
using System.Text;
using System.Collections;

namespace MazeGenerator.MazeGenerator.MazeGeneratorAlgorithm
{
    /// <summary>
    /// Maze generation using Eller's algorithm.
    /// </summary>
    public class MazeGenEller : AbstractImmediateMazeGenerator
    {
        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="width"></param>
        /// <param name="height"></param>
        public MazeGenEller(int width, int height) : base(width, height) { }

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="width"></param>
        /// <param name="height"></param>
        /// <param name="startCellX"></param>
        /// <param name="startCellY"></param>
        public MazeGenEller(int width, int height, int startCellX, int startCellY) : base(width, height, startCellX, startCellY) { }

        List<CellsOfEllerSet> listOfSetsInRow;

        public override void Generate()
        {
            listOfSetsInRow = new List<CellsOfEllerSet>();

            Dictionary<int, int> lastRowCellToSetIndexMap = new Dictionary<int, int>();

            for (int curRowLine = 0; curRowLine < Height; curRowLine++)
            {
                listOfSetsInRow.Clear();

                foreach (int cellNr in lastRowCellToSetIndexMap.Keys)
                {
                    int setId = lastRowCellToSetIndexMap[cellNr];
                    if (containsSet(setId))
                    {
                        addCellToSet(setId, cellNr);
                    }
                    else
                    {
                        CellsOfEllerSet curSet = new CellsOfEllerSet(setId);
                        curSet.AddCell(cellNr);
                        listOfSetsInRow.Add(curSet);
                    }
                }

                int newsetId = 0;
                for (int cellNr = 0; cellNr < Width; cellNr++)
                {
                    if (!cellExistsInAnySet(cellNr))
                    {
                        while (lastRowCellToSetIndexMap.ContainsValue(newsetId))
                        {
                            newsetId++;
                        }

                        CellsOfEllerSet curSet = new CellsOfEllerSet(newsetId);
                        curSet.AddCell(cellNr);
                        listOfSetsInRow.Add(curSet);
                        newsetId++;
                    }
                }

                lastRowCellToSetIndexMap.Clear();

                //printAllSets();
                printRow(curRowLine);
                //System.Diagnostics.Debug.WriteLine("Before merging");

                for (int n = Width - 2; n >= 0; n--)
                {
                    if (curRowLine == Height - 1)
                    {
                        if (!AreCellsInSameSet(n, n + 1))
                        {
                            /* cells n and n+1 are not in same set, so merge both sets n and n+1 into set n */
                            int setId1 = getSetIdForCell(n);
                            int setId2 = getSetIdForCell(n + 1);
                            CellsOfEllerSet set1 = getSet(setId1);
                            CellsOfEllerSet set2 = getSet(setId2);
                            //System.Diagnostics.Debug.WriteLine("Merge set " + setId2 + " into " + setId1);

                            MergeSets(set1, set2);
                            listOfSetsInRow.Remove(set2);
                            RemoveWall(n + 1, curRowLine, MazeDir.WEST);
                        }
                        else
                        {   /* add walls in same sets */
                            bool randFactor = ((rand.Next(2) % 2) == 0);
                            if (randFactor)
                            {
                                SetWall(n + 1, curRowLine, MazeDir.WEST);
                            }
                        }
                    }
                    else
                    {
                        bool randFactor = ((rand.Next(2) % 2) == 0);
                        if (randFactor && (!AreCellsInSameSet(n, n + 1)))
                        {
                            /* cells n and n+1 are not in same set, so merge both sets n and n+1 into set n */
                            int setId1 = getSetIdForCell(n);
                            int setId2 = getSetIdForCell(n+1);
                            CellsOfEllerSet set1 = getSet(setId1);
                            CellsOfEllerSet set2 = getSet(setId2);
                            //System.Diagnostics.Debug.WriteLine("Merge set " + setId2 + " into " + setId1);

                            MergeSets(set1, set2);
                            listOfSetsInRow.Remove(set2);
                            RemoveWall(n + 1, curRowLine, MazeDir.WEST);
                        }
                    }
                }

                //printAllSets();
                printRow(curRowLine);
                //System.Diagnostics.Debug.WriteLine("After merging");

                // update for drawing the maze, no actual part of eller's algorithm, just for maze GUI
                for (int n = 0; n < Width; n++) { VisitCell(n, curRowLine); }

                /* for each set in current row, select randomly a cell and make a vertical passthrough */
                foreach (CellsOfEllerSet curSet in listOfSetsInRow)
                {
                    int elemNum = 0;
                    bool verticalPathExistsOnSet = false;
                    foreach (int cell in curSet)
                    {
                        bool randFactor = ((rand.Next(20) % 2) == 0);
                        if (randFactor || (++elemNum == curSet.Count() && !verticalPathExistsOnSet))
                        {
                            RemoveWall(cell, curRowLine, MazeDir.SOUTH);
                            lastRowCellToSetIndexMap.Add(cell, curSet.GetID());
                            //System.Diagnostics.Debug.WriteLine("Vertical cell: " + cell);
                            verticalPathExistsOnSet = true;
                        }
                    }
                }

            }
        }

        /// <summary>
        /// Merge both given sets into set1.
        /// </summary>
        /// <param name="set1"></param>
        /// <param name="set2"></param>
        private void MergeSets(CellsOfEllerSet set1, CellsOfEllerSet set2)
        {
            foreach (int cellInSet2 in set2)
            {
                set1.AddCell(cellInSet2);
            }
            
        }

        /// <summary>
        /// Check if both given cells are inside the same set.
        /// </summary>
        /// <param name="cellNr1"></param>
        /// <param name="cellNr2"></param>
        /// <returns></returns>
        private bool AreCellsInSameSet(int cellNr1, int cellNr2)
        {
            int set1 = getSetIdForCell(cellNr1);
            int set2 = getSetIdForCell(cellNr2);
            return (set1 == set2);
        }

        //private void printAllSets()
        //{
        //    foreach (CellsOfEllerSet set in listOfSetsInRow)
        //    {
        //        System.Diagnostics.Debug.WriteLine(set.ToString());
        //    }
        //}

        private void printRow(int rowLine)
        {
            //System.Diagnostics.Debug.WriteLine("--- ROW " + rowLine + " ---");

            StringBuilder sbCells = new StringBuilder();
            sbCells.Append("Cells: ");
            for (int cellNr = 0; cellNr < Width; cellNr++)
            {
                sbCells.Append(cellNr + " ");
            }
            //System.Diagnostics.Debug.WriteLine(sbCells.ToString());

            StringBuilder sbSetPerCell = new StringBuilder();
            sbSetPerCell.Append("Sets:  ");
            for (int cellNr = 0; cellNr < Width; cellNr++)
            {
                sbSetPerCell.Append(getSetIdForCell(cellNr) + " ");
            }
            //System.Diagnostics.Debug.WriteLine(sbSetPerCell.ToString());

            //System.Diagnostics.Debug.WriteLine("--- ROW " + rowLine + " End ---");
        }

        public int getSetIdForCell(int cellNr)
        {
            foreach (CellsOfEllerSet set in listOfSetsInRow)
            {
                if (set.ContainsCell(cellNr))
                {
                    return set.GetID();
                }
            }
            throw new InvalidOperationException(cellNr + " not contained in any set");
        }

        public bool cellExistsInAnySet(int cellNr)
        {
            foreach (CellsOfEllerSet set in listOfSetsInRow)
            {
                if (set.ContainsCell(cellNr))
                {
                    return true;
                }
            }
            return false;
        }

        public CellsOfEllerSet getSet(int setID)
        {
            foreach (CellsOfEllerSet set in listOfSetsInRow)
            {
                if (setID == set.GetID())
                {
                    return set;
                }
            }
            throw new InvalidOperationException("Cannot find set with ID " + setID);
        }

        public bool containsSet(int setID)
        {
            foreach (CellsOfEllerSet set in listOfSetsInRow)
            {
                if (setID == set.GetID())
                {
                    return true;
                }
            }
            return false;
        }

        public void addCellToSet(int setId, int cellNr)
        {
            getSet(setId).AddCell(cellNr);
        }
    }
}
