﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Threading;

namespace MazeGenerator.MazeGenerator.StepwiseMazeGeneratorAlgorithm
{
    /// <summary>
    /// Base class to implement a stepwise 
    /// </summary>
    public abstract class AbstractStepwiseMazeGenerator : AbstractMazeGenerator, IDisposable
    {
        /// <summary>
        /// thread which does the calculation
        /// </summary>
        protected Thread thread;

        /// <summary>
        /// Holds the thread state, paused or running
        /// </summary>
        protected bool isPaused;

        /// <summary>
        /// User GUI control where the maze will be drawn. Used to call the MazeChanged event in this tread.
        /// </summary>
        protected DependencyObject parentControl;

        /// <summary>
        /// Event called if the next step of the maze is calculated.
        /// </summary>
        public event EventHandler<MazeChangedEvent> MazeChanged;

        /// <summary>
        /// 
        /// </summary>
        /// <param name="width">width in cells of maze</param>
        /// <param name="height">height in cells of maze</param>
        /// <param name="parentControl">GUI control which will do the painting</param>
        public AbstractStepwiseMazeGenerator(int width, int height, DependencyObject parentControl)
            : this(width, height, parentControl , - 1, -1)
        {
            
        }

        public AbstractStepwiseMazeGenerator(int width, int height, DependencyObject parentControl, int startCellX, int startCellY)
            : base(width, height, startCellX,  startCellY)
        {
            this.parentControl = parentControl;
            isPaused = true;
            thread = new System.Threading.Thread(DoWork); 
        }

        /// <summary>
        /// Start the stepwise calculation.
        /// </summary>
        public void Start()
        {
            thread.Start();
        }

        /// <summary>
        /// Continue to calculate the next step of the maze.
        /// </summary>
        public void Continue()
        {
            isPaused = false;
        }

        /// <summary>
        /// Actual maze calculation routine to be overwritten in subclass.
        /// </summary>
        protected abstract void DoWork();

        /// <summary>
        /// This function shall be called if the next step of the maze was calculated.
        /// This will fire the MazeChanged event and pause the calculation until Continue() is called.
        /// </summary>
        /// <param name="isFinished">flag to specify if the maze is completely calculated.
        /// This value is specified in the event arguments.</param>
        protected void WaitAndNotify(bool isFinished)
        {
            parentControl.Dispatcher.BeginInvoke((ThreadStart)delegate()
            {
                processMazeChangedEvents(isFinished);
            });

            while (isPaused)
                System.Threading.Thread.Sleep(150);
            isPaused = true;
        }

        /// <summary>
        /// Call the event notification of the next maze step.
        /// </summary>
        /// <param name="isFinished"></param>
        protected void processMazeChangedEvents(bool isFinished)
        {
            if (MazeChanged != null)
            {
                MazeChanged(this, new MazeChangedEvent(isFinished));
            }
        }

        public void Dispose()
        {
            thread.Abort();
        }

    }
}
