﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Threading;

namespace MazeGenerator.MazeGenerator.StepwiseMazeGeneratorAlgorithm
{
    /// <summary>
    /// Step-by-step maze calculation of a maze using the Hunt-And-Kill algorithm.
    /// </summary>
    public class StepwiseMazeGenHuntAndKill : AbstractStepwiseMazeGenerator
    {
        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="width"></param>
        /// <param name="height"></param>
        /// <param name="parent"></param>
        public StepwiseMazeGenHuntAndKill(int width, int height, UserControl parent)
            : base(width, height, parent)
        {
             
        }

        public StepwiseMazeGenHuntAndKill(int width, int height, UserControl parent, int startCellX, int startCellY)
            : base(width, height, parent, startCellX, startCellY)
        {

        }

        /// <summary>
        /// Calculates the maze in a separate thread.
        /// </summary>
        protected override void DoWork()
        {
            Tuple<int, int> curCell = Tuple.Create<int, int>(StartCellX, StartCellY);
            VisitCell(StartCellX, StartCellY);

            while (curCell != null)
            {
                walk(curCell);
                curCell = hunt();
            }

            WaitAndNotify(true);
        }

        /// <summary>
        /// This function 'walks' along the cells until it reaches a dead-end cell.
        /// </summary>
        /// <param name="startCell"></param>
        private void walk(Tuple<int, int> startCell)
        {
            Tuple<int, int> curCell = startCell;

            bool deadEnd = false;
            while (!deadEnd)
            {
                deadEnd = true;
                MazeDir[] randMazeDirOrder = GetRandomMazeDirOrder();
                foreach (MazeDir dir in randMazeDirOrder)
                {
                    if (ExistsWall(curCell.Item1, curCell.Item2, dir) &&
                        IsAdjacentCellVisited(curCell.Item1, curCell.Item2, dir) == false)
                    {
                        RemoveWall(curCell.Item1, curCell.Item2, dir);
                        Tuple<int, int> adjCell = Tuple.Create<int, int>(curCell.Item1 + DELTA_X[(ushort)dir],
                                                                         curCell.Item2 + DELTA_Y[(ushort)dir]);
                        VisitCell(adjCell.Item1, adjCell.Item2);
                        curCell = adjCell;
                        deadEnd = false;
                        WaitAndNotify(false);
                        break;
                    }
                }
            }
        }

        /// <summary>
        /// Search for a unvisited cell which has an adjecent cell which is already visited.
        /// </summary>
        /// <returns></returns>
        private Tuple<int, int> hunt()
        {
            for (int x = 0; x < this.Width; x++)
            {
                for (int y = 0; y < this.Height; y++)
                {
                    if (!IsCellVisited(x, y))
                    {
                        /* cell not visited, check if any adjacent cell is already visited */
                        MazeDir[] randMazeDirOrder = GetRandomMazeDirOrder();
                        foreach (MazeDir dir in randMazeDirOrder)
                        {
                            if (IsAdjacentCellInsideMaze(x, y, dir) &&
                                IsAdjacentCellVisited(x, y, dir))
                            {
                                /* found, return the found adjacent cell */
                                return Tuple.Create<int, int>(x + DELTA_X[(ushort)dir], y + DELTA_Y[(ushort)dir]);
                            }
                        }
                    }
                }
            }
            return null;
        }
    }
}
